// Copyright 2007 Paul Milenkovic, 118 Shiloh Dr., Madison, WI 53705 U.S.A.

// Revised to support parallel-filter composition.
//
// Call TransmitNow(NULL,nBuf,nRet) to conduct a dry run to determine the
// number of transmitted samples nRet without calculating filter output or 
// updating the internal input buffer, and then call
// TransmitNow(wave_samp,nBuf,nRet) to go ahead and filter and update the
// internal buffer, knowing what value of nRet to expect.  This feature
// helps keep parallel filters time synchronized.  You can perform
// the dry run on all of the elements of a composite filter to determine
// the allowed nBuf for the actual run that prevents any of the elements
// from getting ahead of the others.
//
// The dry run fetches samples from the source to this filter and tops
// off the internal buffer to a specified amount, but it does not empty
// the buffer, so the nBuf of the actual run can be safely less than the
// nBuf of the dry run, and any extra samples are retained in the internal
// input buffer for the next invocation of TransmitNow().  

#ifndef NonCausalH
#define NonCausalH

#include "SISOFilter.h"

class NonCausal : public SISOFilter
{
private:
    int M_pre, M_pst;
    float src_buf[8192];
    int nsrc; // number of samples stored in src_buf
    int nend; // number of zero-pad samples past input end-of-file
    bool end_of_src; // src EOF plus nend samples zero-pad reached

protected:
    void setM_pre(int M_pre) {this->M_pre = M_pre;}
    void setM_pst(int M_pst) {this->M_pst = M_pst;}
    int  getM_pre() {return M_pre;}
    int  getM_pst() {return M_pst;}

    virtual void SrcSet() 
    {
        SISOFilter::SrcSet();

		// M_pre samples of zero-pad prior to n = 0
        for (int i = 0; i < M_pre; i++)
            src_buf[i] = 0;
        nsrc = M_pre;
		// M_pst samples zero-pad past end of input
        nend = M_pst;
        end_of_src = false;
    }

    virtual void Apply(float* src, float* dst, int nSamp) 
    {
        for (int i = 0; i < nSamp; i++)
            dst[i] = src[M_pre+i];
    }

    int FillSrcBuf(int ndst)
	{
        int nsmp = ndst + M_pre + M_pst;
        if (nsmp > 8192)
            nsmp = 8192;
        nsmp -= nsrc;

        if (nsmp > 0)
        {
            if (!SISOFilter::EndOfSrc())
            {
                int nTransmit;
                SISOFilter::TransmitNow(src_buf+nsrc,nsmp,nTransmit);
                nsrc += nTransmit;
                nsmp -= nTransmit;
            }
            if (SISOFilter::EndOfSrc())
            {   
                if (!end_of_src)
                {
                    // Prepare nend samples zero-pad before raising EndOfSrc()
                    if (nsmp > nend)
                        nsmp = nend;
                    nend -= nsmp;
                    end_of_src = nend == 0;
                }
		        // Returns more zero-pad if invoked past EndOfSrc().
                for (int i = nsrc; i < nsrc + nsmp; i++)
                    src_buf[i] = 0.0f;
                nsrc += nsmp;
            }
        }

		// Allowed output samples based on what is in the buffer, even if 
		// more than requested owing to previous invocation.
        int nRet = nsrc - M_pre - M_pst;
		if (nRet < 0)
			nRet = 0;
        return nRet;
	}

	void ShiftSrcBuf(int ndst)
	{
        for (int i = 0; i < nsrc-ndst; i++)
            src_buf[i] = src_buf[ndst+i];
        nsrc -= ndst;
	}

public:
    NonCausal() : M_pre(0), M_pst(0), nsrc(0), nend(0), end_of_src(true) {}

    virtual bool STDCALL EndOfSrc() {return end_of_src;}
   
    virtual void STDCALL TransmitNow(float* wave_samp, int nBuf, int& nRet)
    {
		nRet = FillSrcBuf(nBuf);
		// Don't return more samples than asked for, even if a previous,
		// wave_samp == NULL invocation had topped off the buffer.
		if (nRet > nBuf)
			nRet = nBuf;

		if (nRet > 0 && wave_samp != NULL) 
		{
            Apply(src_buf,wave_samp,nRet);
			ShiftSrcBuf(nRet);
		}
    }
};

#endif
