// Copyright 2004 Paul Milenkovic, 118 Shiloh Dr., Madison, WI 53705 U.S.A.

#ifndef SISOFilterH
#define SISOFilterH

#include "WaveCpp.h"

class SISOFilter : public IWaveFilterC
{
private:
    IConsoleOutC* console_out;
    IWaveSourceC* src;

    double avg_time;
    int avg_count;
    double t1;

protected:
    IWaveSourceC* getSrc() {return src;}

    void CheckInTime()
    {
        t1 = ElapsedTime();
    }

    void CheckOutTime(int count, const char* str, bool last_time)
    {
        if (count > 0)
        {
            double dt = ElapsedTime() - t1;
            avg_time += 1e9*dt/count;
            avg_count++;
        }
        if (last_time)
        {
            TextOut(str);
            if (avg_count > 0)
                DoubleOut(avg_time/avg_count,10,3);
            OutLn();
        }
    }

    void TextOut(const char* str)
    {
        if (console_out != NULL) console_out->TextOut(str);
    }

    void TextOutLn(const char* str)
    {
        if (console_out != NULL) console_out->TextOutLn(str);
    }

    void OutLn()
    {
        if (console_out != NULL) console_out->OutLn();
    }

    void IntOut(int val, int width)
    {
        if (console_out != NULL) console_out->IntOut(val,width);
    }

    void DoubleOut(double val, int width, int precis)
    {
        if (console_out != NULL) console_out->DoubleOut(val,width,precis);
    }

    double ElapsedTime()
    {
        double t = 0.0;
        if (console_out != NULL)
            console_out->ElapsedTime(t);
        return t;
    }

    virtual void SrcSet()
    {
        IWaveDescriptorC* dsc;
        src->GetDescriptor(&dsc);
        Init(dsc);
        dsc->Release();
    }

    virtual void Init(IWaveDescriptorC* dsc)
    {
        Init(dsc->Get_sample_ms());
    }

    virtual void Init(double sample_ms) {}

public:
    SISOFilter() : 
        console_out(NULL), src(NULL), avg_time(0.0), avg_count(0), t1(0.0) 
    {}

    ~SISOFilter() 
    {
        Dispose();
        Set_console_out(NULL);
    }

    virtual ULONG STDCALL Release() 
    {
        int rc = decRefCount();
        if (rc <= 0) delete this;
        return rc;
    }

    virtual void  STDCALL GetDescriptor(IWaveDescriptorC** dsc) 
    {
        src->GetDescriptor(dsc);
    }

    virtual void  STDCALL TransmitNow(float* wave_samp, int nBuf, int& nRet)
    {
        src->TransmitNow(wave_samp,nBuf,nRet);
    }

    virtual void  STDCALL Dispose() 
    {
        if (src != NULL) 
        {
            src->Dispose();
            src->Release();
            src = NULL;
        }
    }

    virtual bool STDCALL EndOfSrc()
    {
        return (src == NULL) || src->EndOfSrc();
    }

    virtual void  STDCALL EnableInPlace(bool inplace)
    {
        if (src != NULL)
            src->EnableInPlace(inplace);
    }

    virtual void  STDCALL Set_src_list(IWaveSourceC** src_list, int nSrc) 
    {
        Dispose();
        if (nSrc > 0)
            src = src_list[0];
        for (int i = 1; i < nSrc; i++)
            src_list[i]->Release();
        if (nSrc > 0)
            SrcSet();
    }

    virtual void  STDCALL Set_param_list(double* param_list, int nParam) {}

    virtual void  STDCALL Set_console_out(IConsoleOutC* new_console_out)
    {
        if (console_out != NULL)
            console_out->Release();
        console_out = new_console_out;
        if (console_out != NULL)
            console_out->AddRef();
    }

    virtual void  STDCALL Set_num_stream(INumStreamC* num_stream) 
    {
        if (num_stream != NULL)
        {
            // cycle reference count to release object if not used
            num_stream->AddRef();
            num_stream->Release();
        }
    }
};

#endif
