// Copyright 2004 Paul Milenkovic, 118 Shiloh Dr., Madison, WI 53705 U.S.A.

// Factory interface for creating filter and filter parameter dialog 
// objects and supplying interfaces to those objects from a DLL.  
// 
//     IMultiWaveObj -- used by filter dialog objects to apply filters to
//         a TF32 waveform channel -- this interface is supplied by TF32
//         when calling IFilterDialog or IFilterFactory,
//
//     IParamList -- interface to a param list object,
//
//     IFilterDialog -- interface to a filter dialog object that saves and
//         updates filter parameters and applies a filter with those
//         parameters to a TF32 waveform channel.
//
//     IFilterFactory -- interface to a factory object that creates filter
//         or filter dialog objects usable by TF32.

#ifndef WaveBaseH
#define WaveBaseH

#include <string.h>
#include "COMDef.h"
#include "WaveCpp.h"

#define IWaveObj IUnknown

// Interface to the container for one or more waveform channels of the TF32
// waveform display or the TF32Ax ActiveX WavePlotX control.  The filter
// dialog object uses the ApplySISO() method to apply a single-input
// single-output waveform filter to a channel in the display.

struct IMultiWaveObj : public IUnknown
{
    virtual IWaveObj* STDCALL NewWaveObj() = 0;
    virtual IWaveObj* STDCALL AddWaveObj(IWaveObj* link_wave) = 0;
    virtual void      STDCALL EndNewWaveObj() = 0;
    virtual int       STDCALL Get_nchan() = 0;
    virtual IWaveObj* STDCALL GetWaveObj(int ch) = 0;
    virtual int       STDCALL ChfromWaveObj(IWaveObj* wave) = 0;
    virtual bool      STDCALL ApplySISO(IWaveFilterC* f, IWaveObj* src,
        bool newchan) = 0;
};

// Interface to a param list object
//
//     nParams() returns the number of entries in the parameter list,
//     setVal() sets a parameter value selected by index ix --
//         ix ranges from 0 to nParams()-1,
//     getVal() returns a parameter selected by index ix,
//     getValStr() returns a text representation of the parameter,
//     getName() returns a text label for the parameter,
//     getNext() returns the next index in a list of text labels.

class IParamList : public TUnknown
{
public:
    virtual int    STDCALL nParams() = 0;
    virtual int    STDCALL setVal(double val, int ix) = 0;
    virtual double STDCALL getVal(int ix) = 0;
    virtual void   STDCALL getValStr(int ix, char* buf, int nbuf) = 0;
    virtual void   STDCALL getName(int ix, char* buf, int nbuf) = 0;
    virtual int    STDCALL getNext(int ix) = 0;

    string getValStr(int ix)
    {
        char sbuf[512];
        getValStr(ix,sbuf,512);
        return string(sbuf);
    }

    string getName(int ix)
    {
        char sbuf[512];
        getName(ix,sbuf,512);
        return string(sbuf);
    }
};

// Interface to a filter dialog object.
//
//     nParams() returns the number of entries in the parameter list,
//     setVal() sets a parameter value selected by index ix --
//         ix ranges from 0 to nParams()-1,
//     getVal() returns a parameter selected by index ix,
//     ShowDlg() displays a dialog window for editing filter parameters
//         but does not apply the filter,
//     NewChan() returns TRUE if ShowDlg() has selected the newchan button,
//     FilterDlg() displays a dialog window for editing filter parameters
//         and optionally applying the filter to a waveform channel,
//     Apply() applies a filter to a channel in the range
//        1 .. mwI->Get_nchan() -- a ch value of 0 selects the default
//        channel of the display.  Filter output will go to a new channel
//        in the display or overwrite the input channel based on the value
//        of NewChan() returned by this interface.  Combined calls to
//        ShowDlg() followed by Apply() are equivalent to FilterDlg().

class IParamDialog : public TUnknown
{
public:
    virtual int    STDCALL nParams() = 0;
    virtual int    STDCALL setVal(double val, int ix) = 0;
    virtual double STDCALL getVal(int ix) = 0;
    virtual bool   STDCALL ShowDlg() = 0;
};

class IFilterDialog : public IParamDialog
{
public:
    virtual bool   STDCALL NewChan() = 0;
    virtual bool   STDCALL FilterDlg(IMultiWaveObj* mwI, int ch) = 0;
    virtual bool   STDCALL Apply(IMultiWaveObj* mwI, int ch) = 0;
};

// Interface for creating instances of filter or filter dialog objects.
//
// The functions that return objects increment the reference count with
// AddRef() -- you have to call Release() when you are done with one of these
// objects.  These functions return FALSE and return a NULL interface
// reference if they fail.
//
//     SetStr supplies an ASCII string to this object for controlling
//         configuration or other purposes,
//     GetFilter creates and returns a wave filter usable by TF32,
//     GetDialog creates and returns a filter dialog object,
//     GeNamed creates and returns a type of object selected by the
//         character string name -- "WaveFilterC" returns a wave filter
//         object (same as GetFilter), "FilterDialog" returns a filter dialog
//         object (same as GetDialog), other strings may create and return
//         other types of objects depending on future implementations.

class IFilterFactory : public TUnknown
{
public:
    virtual   void STDCALL SetStr(char* str) = 0;
    virtual   bool STDCALL GetFilter(IWaveFilterC** wave_filter) = 0;
    virtual   bool STDCALL GetDialog(IFilterDialog** filter_dlg) = 0;
    virtual   bool STDCALL GetNamed(char* name, IUnknown** named) = 0;
};

#endif
