// Copyright 2004 Paul Milenkovic, 118 Shiloh Dr., Madison, WI 53705 U.S.A.

// C++ interface definitions of wave filters for TF32:
//
//     IConsoleOutC -- interface to a console object supplied by TF32 to a 
//         wave filter for logging or performance measurement,
//
//     INumStreamC -- interface to an ASCII file reader object supplied by 
//         TF32 to a speech synthesis filter,
//
//     IWaveDescriptorC -- waveform-file header information about a waveform 
//         sample stream,
//
//     IWaveSourceC -- provides sequential access to the waveform samples in
//         a channel or from the output of a wave filter,
//
//     IWaveFilterC -- extension to IWaveSourceC that supplies filter output, 
//         can be supplied an IWaveSourceC to accept filter input.
//  
// The low-level C++ interfaces supported by this module are used by C++ DLLs
// (there is also a Pascal version of these interfaces for Delphi Object Pascal), 
// callable from the TF32 stand-alone program or the TF32Ax ActiveX components.  
//
// These interfaces don't make use of any of the "Automation" data types 
// (Variant, Variant arrays, safe arrays) or require implementation
// of IDispatch, support for which introduces considerable overhead in C++.
// Interpreted and or "managed" languages that require Automation data types
// or IDispatch support cannot use these interfaces directly -- the TF32Ax 
// ActiveX library employs wrapper classes to these low-level interfaces to 
// add support for Variant and IDispatch in the situations where it is needed.
//
// Filter objects based on these interfaces obtain waveform input from
// sequential-access waveform stream interface, and the output of each filter
// is in turn supplied as a sequential-access waveform stream interface.  Some
// signal sources may be arrays, other sources may be sequential streams, and
// a filter that relies on a sequential input stream can be readily applied
// to both types of sources.
//
// The IWaveSourceC interface provides sequential access to waveform samples -- 
// the suffix C indicates that this is the low-level C++-compatible interface.  
// A wave source object with the IWaveSourceC interface can be retrieved from a 
// waveform channel to read waveform samples -- multiple wave source objects can 
// be retrieved from the same waveform channel so multiple "viewers" can access 
// the same waveform channel, each inputing samples from a wave source object 
// at its own pace. A single wave source object can also be supplied to a waveform 
// channel to overwrite its contents with a new signal -- this is how signals get 
// read into a waveform channel upon opening a file or how the output of a 
// waveform filter gets written to a channel.  There is a notification mechanism 
// for telling registered viewers of a channel when the contents of a channel 
// have changed, so they can retrieve a new wave source object to access the new 
// signal in the channel.
//
// The IWaveFilterC interface extends the IWaveFilterC interface through C++
// inheritance to allow for waveform input.  An IWaveFilterC object "is" an 
// IWaveSource object, so it can supply waveform samples to write into a channel.  
// An IWaveFilterC object also "has" -- contains -- one or more IWaveSourceC 
// objects to supply input -- while a filter may have multiple inputs, most 
// filters receive a single input and produce a single output.  This system 
// makes it easy to apply filters in cascade -- apply one filter right after 
// the other -- by connecting filters "circus elephant train-style nose-to-tail".  
// The first filter in a cascade can be supplied an IWaveSourceC retrieved from 
// a channel as an input.  That filter "is" an IWaveSourceC and can be supplied 
// as input to the next filter in the chain.  The last filter can be supplied 
// to a channel to overwrite its contents with a new signal.  
//
// TF32 saves the contents of an overwritten channel that is being viewed with 
// wave source objects -- it will keep a copy of any waveform data viewed by 
// wave source objects until all of those wave source objects are released through 
// their reference-counting mechanims.  So it is safe to retrieve the IWaveFilterC 
// object from a channel, supply it as input to a filter or a cascade of filters, 
// and overwrite the channel with the IWaveSourceC of the filter output.
//
// The IWaveDescriptorC interface supplies waveform channel "header" information
// describing sample rate, the number of samples, and other attributes of the 
// signal. The IWaveSourceC and IWaveFilterC interfaces have the GetDescriptor() 
// method to return a reference-counted IWaveDescriptorC object describing 
// attributes of the waveform stream they supply.
//
// IWaveFilterC also has set methods for plug-in interfaces.  The Set_console_obj() 
// and Set_num_stream() methods supply an IConsoleOutC object or an INumStreamC 
// object to the IWaveFilterC object.  The TF32 program will supply an
// IConsoleOutC object to a filter to allow the filter to log text strings and
// to determine elapsed time.  TF32 is able to enable/disable the display of
// such logging information as needed.  The TF32 program will also supply
// an INumStreamC interface to a IWaveFilterC objects used for speech synthesis
// from ASCII parameter files.  INumStreamC is an interface to a reader object
// that provides access to a stream of speech synthesis parameters.


#ifndef WaveCppH
#define WaveCppH

#include <string>
#include "COMDef.h"
using namespace std;

// TF32 will invoke the IWaveFilterC::Set_console_out() method of a wave 
// filter object to supply this interface to a console logging object,
// allowing logging, performance monitoring, and trace debugging.
//
//     TextOut() writes a character string to the console,
//     TextOutLn() same as above but adds a carriage-return line-feed,
//     OutLn() writes a carriage-return line-feed only,
//     IntOut() outputs a string version of an integer of "width" character
//     DoubleOut() outputs a string of a double, "width" characters and "precis"
//         numerals after the decimal point,
//     ElabsedTime() returns time in ms in variable etime.

struct IConsoleOutC : public IUnknown
{
    virtual void   STDCALL TextOut(const char* str) = 0;
    virtual void   STDCALL TextOutLn(const char* str) = 0;
    virtual void   STDCALL OutLn() = 0;
    virtual void   STDCALL IntOut(int val, int width) = 0;
    virtual void   STDCALL DoubleOut(double val, int width, int precis) = 0;
    virtual void   STDCALL ElapsedTime(double& etime) = 0;
};


// TF32 supplies this interface to an ASCII file reader object to supply
// a parameter streams to a speech synthesis filter.
//
//     Open() to open a file with the supplied file name,
//     GetSingle() reads the file for a type float,
//     GetDouble() reads the file for a type double,
//     GetLine() reads past the next line break in the ASCII file,
//     Close() closes any open file.
//
// All of the functions returning type bool return FALSE on failure.

struct INumStreamC : public TUnknown
{
    virtual bool   STDCALL Open(char* fname) = 0;
    virtual bool   STDCALL GetSingle(float& val) = 0;
    virtual bool   STDCALL GetDouble(double& val) = 0;
    virtual bool   STDCALL GetLine() = 0;
    virtual void   STDCALL Close() = 0;
};


// Interface to a wave descriptor object describing waveform channel
// "header" information.  There are Get/Set functions for properties
//
//     msamp is the total number of waveform samples in a channel,
//     sample_ms is the time in milliseconds betweenw waveform samples,
//     start_ms is the start time of the stream -- usually 0 ms,
//     range is the peak-to-peak range of signal values in the 
//         designated units, usually 20 for 20 Volts peak-to-peak
//         unless changed by a signal calibration,
//     offset is the amount, in designated units, subtracted from
//         the signal to correct for DC offset -- usually 0.0 unless
//         a signal has been calibrated and null-offset corrected,
//     res is the number of bits used to represent waveform samples,
//        typically 15 bits for waveform channels to allow 1 bit
//        headroom for signal overflow, 16 bits for F0 trace,
//     uname is the units name, typically Volts unless calibrated,
//     fname is the file from which the waveform was opened,
//     pname is the plot name describing a configuration of channels
//         for TF32 channel presets (requires Lab Automation level),
//     cname is a descriptive name of the channel: audio, F0.

struct IWaveDescriptorC : public IUnknown
{
    virtual int    STDCALL Get_msamp() = 0;
    virtual void   STDCALL Set_msamp(int Value) = 0;
    virtual void   STDCALL Get_sample_ms(double& Ret) = 0;
    virtual void   STDCALL Set_sample_ms(double Value) = 0;
    virtual void   STDCALL Get_start_ms(double& Ret) = 0;
    virtual void   STDCALL Set_start_ms(double Value) = 0;
    virtual void   STDCALL Get_range(double& Ret) = 0;
    virtual void   STDCALL Set_range(double Value) = 0;
    virtual void   STDCALL Get_offset(double& Ret) = 0;
    virtual void   STDCALL Set_offset(double Value) = 0;
    virtual int    STDCALL Get_res() = 0;
    virtual void   STDCALL Set_res(int Value) = 0;
    virtual void   STDCALL Get_uname(char* Value, int nBuf) = 0;
    virtual void   STDCALL Set_uname(const char* Value) = 0;
    virtual void   STDCALL Get_fname(char* Value, int nBuf) = 0;
    virtual void   STDCALL Set_fname(const char* Value) = 0;
    virtual void   STDCALL Get_cname(char* Value, int nBuf) = 0;
    virtual void   STDCALL Set_cname(const char* Value) = 0;
    virtual void   STDCALL Get_pname(char* Value, int nBuf) = 0;
    virtual void   STDCALL Set_pname(const char* Value) = 0;

    double Get_sample_ms()
    {
        double sample_ms;
        Get_sample_ms(sample_ms);
        return sample_ms;
    }

    double Get_start_ms()
    {
        double start_ms;
        Get_start_ms(start_ms);
        return start_ms;
    }

    double Get_range()
    {
        double range;
        Get_range(range);
        return range;
    }

    double Get_offset()
    {
        double offset;
        Get_offset(offset);
        return offset;
    }

    string Get_uname()
    {
        char buf[512];
        Get_uname(buf,512);
        return string(buf);
    }

    string Get_fname()
    {
        char buf[512];
        Get_fname(buf,512);
        return string(buf);
    }

    string Get_cname()
    {
        char buf[512];
        Get_cname(buf,512);
        return string(buf);
    }

    string Get_pname()
    {
        char buf[512];
        Get_pname(buf,512);
        return string(buf);
    }
};


// Interface to a wave source object providing sequential access to waveform 
// samples in a channel or supplyin a channel with new waveform samples.
//
//     GetDescriptor() returns an interface to the wave descriptor object:
//         the returned interface has had AddRef() called to increase the
//         reference count -- you must call Release() when you are done with
//         the interface,
//     TransmitNow() retrieves up to nBuf samples into the C++ array wave_samp
//         (make sure wave_samp is at least nBuf samples long), nRet is the
//         actual number of samples retrieved,
//     Dispose() releases the resources held by the wave source object prior
//        to its eventual deletion by the reference-counting process,
//     EndOfSrc() indicates there are no more waveform samples -- additional
//         calls to TransmitNow() may return zero-valued samples or no 
//         samples depending on the wave source object,
//     EnableInPlace() receives a signal that this source is overwriting a
//         waveform channel.

class IWaveSourceC : public TUnknown
{
public:
    virtual void  STDCALL GetDescriptor(IWaveDescriptorC** dsc) = 0;
    virtual void  STDCALL TransmitNow(float* wave_samp, int nBuf, int& nRet) = 0;
    virtual void  STDCALL Dispose() = 0;
    virtual bool  STDCALL EndOfSrc() = 0;
    virtual void  STDCALL EnableInPlace(bool inplace) = 0;
};


// Interface to a wave filter object -- note the assignement of a GUID to
//     IWaveFilterC.
//
//     Set_src_list() supplies one or more IWaveSourceC objects as filter
//         inputs -- each element of the C++ array must have its reference
//         count incremented with AddRef(), and the filter object must
//         invoke Release() on each element of the array when it is done with it.
//     Set_param_list() supplies parameter values as an array of doubles,
//     Set_console_out() is called by TF32 to supply an IConsoleOutC object
//         for logging, performance monitoring, or trace debugging,
//     Set_num_stream() is called by TF32 to supply a file reader object of
//         and ASCII parameter stream used in speech waveform synthesis.
//
// The implementation of the GetDescriptor() method, needs to call the 
// GetDescriptor() method of its input wave source object and copy the input 
// descriptor to the output parameter.  An upsampling filter needs to change 
// properties of that descriptor before supplying it as output; it uses Get and 
// Set functions of the IWaveDescriptorC interface to reduce  the sample_ms value
// and reciprocally increase the msamp property.  That way a descriptor object 
// can be retrieved from the input and forwarded to the output if the waveform 
// properties do not change, or the descriptor properties may be modified to 
// reflect waveform changes between filter input and output.
//
// The implementation of EnableInPlace() needs to forward this call to the input
// IWaveSourceC object with the value of the input parameter inplace.  When 
// EnableInPlace() gets invoked on a wave source supplying filter input at the
// start of a filter chain, TF32 determines whether the filter output can share the
// same memory buffer as the filter input, saving on memory for long signals and 
// repeated filter operations.  You need to forward this call as 
// EnableInPlace(FALSE) if the filter is unable to safely overwrite the waveform 
// channel input buffer with the filter output.

class IWaveFilterC : public IWaveSourceC
{
public:
    virtual HRESULT STDCALL QueryInterface(const struct _GUID& guid, void ** pvObj) 
    {
        _GUID IID_IWaveFilterC = 
            {0x3BD1AA82,0xD4ED,0x49A1,{0xB4,0x7A,0x65,0x6D,0x23,0x66,0x89,0xBF}};
        if (IsEqualGUID((_GUID*) &guid,(_GUID*) &IID_IWaveFilterC))
        {
            AddRef();
            *pvObj = this;
            return S_OK;
        } 
        else 
            return TUnknown::QueryInterface(guid,pvObj);
    }

    virtual void  STDCALL Set_src_list(IWaveSourceC** src_list, int nSrc) = 0;
    virtual void  STDCALL Set_param_list(double* param_list, int nParam) = 0;
    virtual void  STDCALL Set_console_out(IConsoleOutC* console_out) = 0;
    virtual void  STDCALL Set_num_stream(INumStreamC* num_stream) = 0;
};

#endif
