// Copyright 2007 Paul Milenkovic, 118 Shiloh Dr., Madison, WI 53705 U.S.A.

#ifndef BeamDataH
#define BeamDataH

#include <math.h>
#include "AddTrack.h"
#include "NumStream.h"

#define mpel 16
#define epoch 6.866
#define epoch2 0.006866
#define mpal 4096

class BeamData : public AddTrack 
{
private:
    double sample_ms;

    bool valid_epoch;
    double t_epoch;
    int i_epoch;
	int x1,y1,x2,y2;

    double pel_val[mpel];
	double pel_val_pre[mpel];
	double pel_val_ppr[mpel];
	double pel_out;
	double pel_num1, pel_num2;
	double pel_posx5;
	double pel_posx_old5;
	double pel_posy5;
	double pel_posy_old5;
	double pel_vel;
	double pel_vel_old;
	double pel_mode;

	double pal_x[mpal];
	double pal_y[mpal];
	int npal;

    bool GetEpoch()
    {
        bool result = GetLine();
		t_epoch += epoch;
        i_epoch = int(t_epoch/sample_ms);
        return result;
    }

	double dist(double x0, double y0, double x1, double y1)
	{
		return sqrt((x1 - x0)*(x1 - x0) + (y1 - y0)*(y1 - y0));
	}

	double offset_angle(double x0, double y0, double x1, double y1)
	{
		//double sign;
		//double diff1 = y1-y0;
		//double diff2 = x1-x0;
		//if (diff1 >= 0)
		//	sign = 1.0;
		//else
		//	sign = -1.0;
		//if (diff2 = 0)
		//	return PI/4.0 * sign;
		//else
			return atan2((y1-y0),(x1-x0));
			//* sign;
	}

	void ComputePelOut()
	{
		pel_out = 0.0;
		switch ( (int) pel_mode )
		{ 
			case 1 : 
				//distance between pellet 1 and pellet 2
				if (pel_val[4] < 1000000.0 &&
					pel_val[5] < 1000000.0 &&
					pel_val[6] < 1000000.0 &&
					pel_val[7] < 1000000.0
					)
				    pel_out = dist(pel_val[4], pel_val[5], pel_val[6], pel_val[7])/1000.0;

				break;
			case 2 :
				//r-tip-y
				if (pel_val[4] < 1000000.0 &&
					pel_val[5] < 1000000.0 &&
					pel_val[6] < 1000000.0 &&
					pel_val[7] < 1000000.0 &&
					pel_val[8] < 1000000.0 &&
					pel_val[9] < 1000000.0 
					)
				{
					// Need to set pel_out to computed value, need to check what range
					// and units pel_out has -- modify GetDescriptor() to act on pel_mode
					// as required.

					double d12 = dist(pel_val[4], pel_val[5], pel_val[6], pel_val[7]);

					double offs1 = offset_angle(pel_val[4], pel_val[5], pel_val[6], pel_val[7]);
					double offs2 = offset_angle(pel_val[6], pel_val[7], pel_val[8], pel_val[9]);
					double theta = offs2 - offs1;
					pel_out = d12 * sin(theta)/1000.0;
				}
				break;
			case 3 : 
				//y-velocity of tongue pellet 1
				if (pel_val[5] < 1000000.0 && pel_val_pre[5] < 1000000.0)
				    pel_out = (pel_val[5] - pel_val_pre[5])/epoch;
				break;
			case 4 : 
				//y-acceleration of tongue pellet 1
				if (pel_val[5] < 1000000.0 &&
					pel_val_pre[5] < 1000000.0 &&
					pel_val_ppr[5] < 1000000.0)
				{
					double vel = (pel_val[5] - pel_val_pre[5])/epoch;
					double vel_pre = (pel_val_pre[5] - pel_val_ppr[5])/epoch;
					pel_out = 1000.0*(vel - vel_pre);
				}
				break;
			case 5 : 
				//absolute velocity (Paul's)
				if (pel_val[4] < 1000000.0 &&
					pel_val[5] < 1000000.0 
					)
				{
					pel_posx5 = pel_val[4]/1000.0;
					pel_posy5 = pel_val[5]/1000.0;
				    pel_out = dist(pel_posx5, pel_posy5, pel_posx_old5, pel_posy_old5)/ epoch;
					pel_posx_old5 = pel_posx5;
					pel_posy_old5 = pel_posy5;
				}
				break;
			case 6 : 
				//absolute acceleration (Paul's)
				if (pel_val[4] < 1000000.0 &&
					pel_val[5] < 1000000.0 
					)
				{
					pel_posx5 = pel_val[4]/1000.0;
					pel_posy5 = pel_val[5]/1000.0;
				    pel_vel = 1000 * dist(pel_posx5, pel_posy5, pel_posx_old5, pel_posy_old5)/ epoch;
					pel_posx_old5 = pel_posx5;
					pel_posy_old5 = pel_posy5;
					pel_out = (pel_vel - pel_vel_old) / epoch;
					pel_vel_old = pel_vel;
				}
				break;
			case 7 : 
				//distance between two pellets
				if (pel_num1 > 0.0 &&
					pel_num2 > 0.0 &&
					pel_val[x1] < 1000000.0 &&
					pel_val[y1] < 1000000.0 &&
					pel_val[x2] < 1000000.0 &&
					pel_val[y2] < 1000000.0 
					)
				{
					x1 = 2*pel_num1-2;
					y1 = 2*pel_num1-1;
					x2 = 2*pel_num2-2;
					y2 = 2*pel_num2-1;
				    pel_out = dist(pel_val[x1], pel_val[y1], pel_val[x2], pel_val[y2])/1000.0;
				}
				break;
			case 8 : 
				//distance between tongue tip (pellet 1) and palate
				if (pel_val[4] < 1000000.0 &&
					pel_val[5] < 1000000.0
					)
				{
					double val_min = 1000000.0;
					int ind_min = 0;
					double dist_pal = 1000000.0;
					int i = 0;
					int lim = fmin(npal-1, 2048);
					while (i<lim)
					{
						dist_pal = dist(pel_val[4], pel_val[5], pal_x[i], pal_y[i]);
						if (dist_pal < val_min){
							val_min = dist_pal;
							ind_min = i;
						}
						i++;
					}
					if (pel_num1 == 1)
						pel_out = float(ind_min + 1);
					else
						pel_out = val_min/1000.0;
				}
				break;

		}
	}

    bool GetParams()
    {
		double pel_inc[mpel];

		int i = 0;
		//read in first mpel numbers from pellet data line
		while (i < mpel && GetNum(pel_inc[i])) i++;

		if (i < mpel)
			return false;
		else {
			for (i = 0; i < mpel; i++)
			    pel_val[i] += pel_inc[i];
			ComputePelOut();
			for (i = 0; i < mpel; i++)
				pel_val_ppr[i] = pel_val_pre[i];
			for (i = 0; i < mpel; i++)
				pel_val_pre[i] = pel_val[i];
			return true;
		}
    }

    void PutPulse(float* sig_buf, int ipulse, int fpulse, int nbuf)
    {
		if (ipulse < 0) ipulse = 0;
		if (fpulse > nbuf) fpulse = nbuf;
		for (int i = ipulse; i < fpulse; i++)
//            sig_buf[i] = (float) (pel_out/1000.0);
            sig_buf[i] = (float) (pel_out);
    }

protected:
    int Get_mem() {return 4096;}

    void Init(IWaveDescriptorC* dsc)
    {
        SISOFilter::Init(dsc);

        npal = 0;
	    NumStream* ns = new NumStream();
		if (ns->OpenFromString(SetFname(dsc->Get_fname(),"pal.dat")))
		{
			while (npal < mpal &&
				   ns->GetLine() && 
				   ns->GetDouble(pal_x[npal]) &&
				   ns->GetDouble(pal_y[npal])
			)
				npal++;
			ns->Close();
            valid_epoch = Open(SetExt(dsc->Get_fname(),".xyd")) && GetEpoch();
		}
		delete ns;
    }

    void Init(double sample_ms)
    {
        this->sample_ms = sample_ms;
		t_epoch = -6.866*1.5;
		for (int i = 0; i < mpel; i++)
		    pel_val[i] = 0.0;
		pel_out = 0.0;
    }

    void Apply(float sig_buf[], int nbuf, int npos)
    {
		if (i_epoch < npos)
			i_epoch = npos;
		for (int i = i_epoch-npos; i < nbuf; i++)
			sig_buf[i] = 0.0f;

		while (valid_epoch && i_epoch-npos < nbuf - 1024) 
        {
            if (GetParams()) {
				int prev_i_epoch = i_epoch;
                valid_epoch = GetEpoch();
                PutPulse(sig_buf,prev_i_epoch-npos,i_epoch-npos,nbuf);
			} else 
                valid_epoch = false;
		}
    }

public:
	BeamData() : valid_epoch(false), pel_mode(0.0)
	{
		for (int i = 0; i < mpel; i++) 
			pel_val_pre[i] = 1000000.0;
		for (int i = 0; i < mpel; i++) 
			pel_val_ppr[i] = 1000000.0;
	}

	virtual void  STDCALL GetDescriptor(IWaveDescriptorC** dsc) 
    {
		AddTrack::GetDescriptor(dsc);
		if (pel_mode == 4.0)
		{
			(*dsc)->Set_uname("mm/s^2");
			(*dsc)->Set_range(50000.0);
		}
		else		
		if (pel_mode == 3.0)
		{
			(*dsc)->Set_uname("mm/s");
			(*dsc)->Set_range(500.0);
		}
		else		
		if (pel_mode == 8.0)
		{
			(*dsc)->Set_uname("mm/s");
			(*dsc)->Set_range(100.0);
		}
		else
		{
			(*dsc)->Set_uname("mm");
			(*dsc)->Set_range(150.0);
		}
		/*
		if (pel_mode == 2.0)
		{
			(*dsc)->Set_uname("deg");
			(*dsc)->Set_range(1000.0);
		}
		*/
    }

	virtual void  STDCALL Set_param_list(double* param_list, int nParam) 
	{
		pel_num1 = 0;
		pel_num2 = 0;
		if (nParam > 0)
		    pel_mode = (int) param_list[0];
		if (nParam > 1) 
			pel_num1 = (int) param_list[1];
		if (nParam > 2) 
			pel_num2 = (int) param_list[2];
	}
};

#endif
