// Copyright 2004 Paul Milenkovic, 118 Shiloh Dr., Madison, WI 53705 U.S.A.

#ifndef ResampleH
#define ResampleH

#include "SISOFilter.h"
#include <math.h>

class Resample : public SISOFilter
{
private:
    int M_pre, M_pst; // pre: look-behind; pst: look-ahead.
    int nUp, nDn, iiU, iiD, iUp, iDn; 
	int	nsrc, nend;
	bool end_of_src;
	float src_buf[8192];
    float fcoef[4097];
	int ncoef;

protected:
	void setM_pre(int M_pre) {this->M_pre = M_pre;}
	void setM_pst(int M_pst) {this->M_pst = M_pst;}
	int  getM_pre() {return M_pre;}
	int  getM_pst() {return M_pst;}
	int  getnUp() {return nUp;}
	int  getnDn() {return nDn;}

	virtual void CalcCoef(float* fcoef, int& ncoef)
	{
		ncoef = 1;
		fcoef[0] = 1.0;
	}

	virtual void SrcSet() 
	{
		SISOFilter::SrcSet();

		CalcCoef(fcoef,ncoef);

		nsrc = M_pre;
		for (int i = 0; i < nsrc; i++)
			src_buf[i] = 0;

		iUp = iiU;
		iDn = iiD;

		nend = M_pst;
		end_of_src = false;
	}

	virtual void Apply(float* src, float* dst, int mSrc, int mDst, int& nSrc, int& nDst) 
	{
		nSrc = 0;
        nDst = 0;
        while (nSrc < mSrc && nDst < mDst)
        {
			// iDn is downsampling-selection index,
			// iUp is polyphase-filter offset index used in upsampling.

            if (iDn == 0)
            {
                int n = nSrc + M_pre + M_pst;
                int k = ncoef - 1 - iUp;
				if (iUp < 0)
					k -= nUp;
                double sum = 0.0;
				while (k >= 0)
				{
					sum += src[n]*fcoef[k];
					n--;
					k -= nUp;
				}
				dst[nDst] = (float) sum;
				// Advance output index for each downsampled output value.
                nDst++;
            }

            iUp++;
            if (iUp >= nUp)
            {
                iUp = 0;
				// Advance input index upon cycling polyphase index.
                nSrc++;
            }

            iDn++;
            if (iDn >= nDn)
                iDn = 0;
        }
	}

public:
	Resample() : M_pre(0), M_pst(0), nUp(1), nDn(1), iiU(0), iiD(0), end_of_src(true)
    {
		Set_param_list(NULL,0);
    }

	virtual void  STDCALL Set_param_list(double* param_list, int nParam) 
    {
		if (nParam > 0) nDn = (int) param_list[0];
        if (nParam > 1) nUp = (int) param_list[1];
		if (nParam > 2) iiD = (int) param_list[2];
        if (nParam > 3) iiU = (int) param_list[3];
		if (nDn < 1) nDn = 1;
		if (nUp < 1) nUp = 1;
    }

	virtual bool STDCALL EndOfSrc() 
	{
		return end_of_src;
	}

	virtual void  STDCALL GetDescriptor(IWaveDescriptorC** dsc) 
	{
        SISOFilter::GetDescriptor(dsc);
        (*dsc)->Set_sample_ms(((*dsc)->Get_sample_ms()*nDn)/nUp);
        (*dsc)->Set_msamp(((*dsc)->Get_msamp()*nUp)/nDn);
	}

	virtual void  STDCALL TransmitNow(float* wave_samp, int nBuf, int& nRet)
	{	
        int nSamp = (nBuf*nDn + nUp - 1)/nUp;
		if (nSamp > 8192-M_pre-M_pst)
			nSamp = 8192-M_pre-M_pst;

        int nReq = nSamp + M_pre + M_pst - nsrc;
        if (!SISOFilter::EndOfSrc() && nReq > 0)
        {
            int nTransmit;
            SISOFilter::TransmitNow(src_buf+nsrc,nReq,nTransmit);
            nsrc += nTransmit;
            nReq -= nTransmit;
        }
        if (SISOFilter::EndOfSrc() && nReq > 0)
        {
            if (!end_of_src)
            {
                if (nReq > nend)
                    nReq = nend;
                nend -= nReq;
            }
            for (int i = nsrc; i < nsrc + nReq; i++)
                src_buf[i] = 0.0;
            nsrc += nReq;
        }
            
		nRet = 0;
		if (nsrc - M_pre - M_pst > 0)
		{
            int nAdv;
		    Apply(src_buf,wave_samp,nsrc-M_pre-M_pst,nBuf,nAdv,nRet);

		    // Shift src_buf.
			for (int i = 0; i < nsrc - nAdv; i++)
				src_buf[i] = src_buf[nAdv+i];
			nsrc -= nAdv;
		}

		end_of_src = nend <= 0 && nsrc - M_pre - M_pst <= 0 && SISOFilter::EndOfSrc();
	}
};

#endif
